import logging
import sqlite3 as sq
from datetime import datetime
from enum import IntEnum, IntFlag
from os.path import isfile
from typing import NamedTuple

experiment_id = int
camera_id = int
stick_id = int
image_id = int
stick_image_id = int


class BoundingBox(NamedTuple):
    x_top: int
    y_top: int
    x_bottom: int
    y_bottom: int

    def __str__(self) -> str:
        return f"({self.x_top}, {self.y_top}, {self.x_bottom}, {self.y_bottom})"

    def __conform__(self, protocol) -> str:
        if protocol is sq.PrepareProtocol:
            return self.__str__()


class ImageUserAction(IntEnum):
    NO_ACTION = 0
    DISCARDED = 1
    APPROVED = 2


class ImageProcessedState(IntEnum):
    UNPROCESSED = 0
    PROCESSED = 1
    SKIPPED = 2


class StickUserAction(IntFlag):
    NO_ACTION = 0
    CREATED = 1
    DISCARDED = 2
    EDITED = 4


class Database():

    def __init__(self, database_path: str) -> None:
        self.con = None
        self.cur = None
        self._create_connection(database_path)
        self._create_tables()

    def _create_connection(self, database_path: str) -> None:
        try:
            if isfile(database_path):
                logging.info(f'Opening database {database_path} ...')
            else:
                logging.info(
                    f'Database {database_path} not found. Creating a new one...')

            self.con = sq.connect(database_path)
            self.cur = self.con.cursor()

            logging.info('...database opened.')
        except sq.Error as e:
            logging.error(e)

    def _create_tables(self):
        try:
            self.cur.execute('PRAGMA foreign_keys = ON;')
            self.con.commit()
            self.cur.executescript('''
                BEGIN;
                CREATE TABLE IF NOT EXISTS Experiments(
                    id INTEGER PRIMARY KEY,
                    location_name TEXT,
                    year YEAR,
                    start TEXT,
                    end TEXT,
                    latitude REAL,
                    longitude REAL
                );
                CREATE TABLE IF NOT EXISTS Cameras(
                    id INTEGER PRIMARY KEY,
                    experiment_id INTEGER NOT NULL,
                    camera_name TEXT NOT NULL,
                    FOREIGN KEY (experiment_id)
                        REFERENCES Experiments (id)
                        ON UPDATE CASCADE
                        ON DELETE CASCADE
                );
                CREATE TABLE IF NOT EXISTS Images(
                    id INTEGER PRIMARY KEY,
                    camera_id INTEGER NOT NULL,
                    file_path TEXT,
                    processed_state INT,
                    datetime TEXT,
                    snow_present INTEGER,
                    is_day INTEGER NOT NULL,
                    reciprocal_image_id TEXT,
                    user_action INT,
                    FOREIGN KEY (camera_id)
                        REFERENCES Cameras (id)
                        ON UPDATE CASCADE
                        ON DELETE CASCADE
                );
                CREATE TABLE IF NOT EXISTS Sticks(
                    id INTEGER PRIMARY KEY,
                    experiment_id INTEGER NOT NULL,
                    camera_id INTEGER NOT NULL,
                    stick_name TEXT,
                    length_cm INT,
                    primary_camera INTEGER NOT NULL,
                    etalon_bounding_box TEXT,
                    etalon_image_id INTEGER NOT NULL,
                    FOREIGN KEY (experiment_id)
                        REFERENCES Experiments (id)
                        ON UPDATE CASCADE
                        ON DELETE CASCADE,
                    FOREIGN KEY (camera_id)
                        REFERENCES Cameras (id)
                        ON UPDATE CASCADE
                        ON DELETE CASCADE
                );
                CREATE TABLE IF NOT EXISTS Stick_Images(
                    id INTEGER PRIMARY KEY,
                    stick_id INTEGER NOT NULL,
                    camera_id INTEGER NOT NULL,
                    image_id INTEGER NOT NULL,
                    bounding_box TEXT,
                    percent_covered REAL,
                    visible INTEGER NOT NULL,
                    user_action INT,
                    FOREIGN KEY (stick_id)
                        REFERENCES Sticks (id)
                        ON UPDATE CASCADE
                        ON DELETE CASCADE,
                    FOREIGN KEY (camera_id)
                        REFERENCES Cameras (id)
                        ON UPDATE CASCADE
                        ON DELETE CASCADE,
                    FOREIGN KEY (image_id)
                        REFERENCES Images (id)
                        ON UPDATE CASCADE
                        ON DELETE CASCADE
                );
                COMMIT;
                ''')
        except sq.Error as e:
            logging.error(e)

    def close(self):
        self.con.close()

    def insert_experiment(self,
                          location_name: str,
                          year: int,
                          start: datetime,
                          end: datetime,
                          latitude: float,
                          longitude: float) -> experiment_id:
        try:
            self.cur.execute('''
                             INSERT INTO Experiments (location_name,
                                                      year,
                                                      start,
                                                      end,
                                                      latitude,
                                                      longitude)
                                 VALUES (:location_name,
                                         :year,
                                         :start,
                                         :end,
                                         :latitude,
                                         :longitude);
                             ''',
                             {'location_name': location_name,
                              'year': year,
                              'start': start,
                              'end': end,
                              'latitude': latitude,
                              'longitude': longitude})
            self.con.commit()
        except sq.Error as e:
            logging.error(e)
        return self.cur.lastrowid

    def insert_camera(self,
                      experiment_id: experiment_id,
                      camera_name: str) -> camera_id:
        try:
            self.cur.execute('''
                             INSERT INTO Cameras (experiment_id,
                                                  camera_name)
                                 VALUES (:experiment_id,
                                         :camera_name);
                             ''',
                             {'experiment_id': experiment_id,
                              'camera_name': camera_name})
            self.con.commit()
        except sq.Error as e:
            logging.error(e)
            return None
        return self.cur.lastrowid

    def insert_image(self,
                     camera_id: camera_id,
                     file_path: str,
                     processed_state: ImageProcessedState,
                     datetime: datetime,
                     snow_present: bool | None,
                     is_day: bool | None,
                     reciprocal_image_id: image_id | None,
                     user_action: ImageUserAction | None):
        try:
            self.cur.execute('''
                             INSERT INTO Images (camera_id,
                                                 file_path,
                                                 processed_state,
                                                 datetime,
                                                 snow_present,
                                                 is_day,
                                                 reciprocal_image_id,
                                                 user_action)
                                 VALUES (:camera_id,
                                         :file_path,
                                         :processed_state,
                                         :datetime,
                                         :snow_present,
                                         :is_day,
                                         :reciprocal_image_id,
                                         :user_action);
                             ''',
                             {'camera_id': camera_id,
                              'file_path': file_path,
                              'processed_state': processed_state,
                              'datetime': datetime,
                              'snow_present': snow_present,
                              'is_day': is_day,
                              'reciprocal_image_id': reciprocal_image_id,
                              'user_action': user_action})
            self.con.commit()
        except sq.Error as e:
            logging.error(e)
            return None
        return self.cur.lastrowid

    def insert_stick(self,
                     experiment_id: experiment_id,
                     camera_id: camera_id,
                     stick_name: str,
                     length_cm: int,
                     primary_camera: camera_id,
                     etalon_bounding_box: BoundingBox,
                     etalon_image_id: image_id) -> stick_id:
        try:
            self.cur.execute('''
                             INSERT INTO Sticks (experiment_id,
                                                 camera_id,
                                                 stick_name,
                                                 length_cm,
                                                 primary_camera,
                                                 etalon_bounding_box,
                                                 etalon_image_id)
                                 VALUES (:experiment_id,
                                         :camera_id,
                                         :stick_name,
                                         :length_cm,
                                         :primary_camera,
                                         :etalon_bounding_box,
                                         :etalon_image_id);
                             ''',
                             {'experiment_id': experiment_id,
                              'camera_id': camera_id,
                              'stick_name': stick_name,
                              'length_cm': length_cm,
                              'primary_camera': primary_camera,
                              'etalon_bounding_box': etalon_bounding_box,
                              'etalon_image_id': etalon_image_id})
            self.con.commit()
        except sq.Error as e:
            logging.error(e)
            return None
        return self.cur.lastrowid

    def insert_stick_image(self,
                           stick_id: stick_id,
                           camera_id: camera_id,
                           image_id: image_id,
                           bounding_box: BoundingBox,
                           percent_covered: float,
                           visible: bool,
                           user_action: StickUserAction) -> stick_image_id:
        try:
            self.cur.execute('''
                             INSERT INTO Stick_Images (stick_id,
                                                       camera_id,
                                                       image_id,
                                                       bounding_box,
                                                       percent_covered,
                                                       visible,
                                                       user_action)
                                 VALUES (:stick_id,
                                         :camera_id,
                                         :image_id,
                                         :bounding_box,
                                         :percent_covered,
                                         :visible,
                                         :user_action);
                             ''',
                             {'stick_id': stick_id,
                              'camera_id': camera_id,
                              'image_id': image_id,
                              'bounding_box': bounding_box,
                              'percent_covered': percent_covered,
                              'visible': visible,
                              'user_action': user_action})
            self.con.commit()
        except sq.Error as e:
            logging.error(e)
            return None
        return self.cur.lastrowid
